#===================================================================================================
# get_refresh_token.py
#
# Simba Technologies Inc.
# Copyright (C) 2023 Simba Technologies Incorporated
#
# Generate Google oauth 2.0 refresh token for using in connection string.
# This script will open a browser for user to log in and will listen to 
# a port for getting the authorization code.
#
#
# Users can change the scope or client id and client secret if necessary.
#
#===================================================================================================

import socket
import sys
import http.client
import json
import base64
import os
import re
import hashlib
import webbrowser
from html.parser import HTMLParser

if sys.version_info.major < 3 :
    print('This script requires python 3. Please install python 3 and rerun the script.')
    sys.exit()

# specify the request requirments
#HOST = '127.0.0.1' always localhost
SCOPE = 'https://www.googleapis.com/auth/bigquery'
''' In case drive scope was needed.'''
#SCOPE = 'https://www.googleapis.com/auth/bigquery https://www.googleapis.com/auth/drive'
CLIENT_ID = '977385342095.apps.googleusercontent.com'
CLIENT_SECRET = 'wbER7576mc_1YOII0dGk7jEE'

class UserAccountAuthentication:
    def __init__(self, client_id,client_secret,scope):
        self.clientId = client_id
        self.clientSecret = client_secret
        self.scope = scope
    
    def ExtractCode(self, data):
        try:
            codeStart = data.index('code=') + 5
            codeEnd = data.index( '&', codeStart )
            if codeStart < codeEnd:
                code = data[codeStart:codeEnd]
                if len(code)<10:
                    return ""
                return code
            return ""
        except ValueError:
            return ""

    def generate_code_verifier_and_challenge(self):
        code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode('utf-8')
        code_verifier = re.sub('[^a-zA-Z0-9]+', '', code_verifier)

        code_challenge = hashlib.sha256(code_verifier.encode('utf-8')).digest()
        code_challenge = base64.urlsafe_b64encode(code_challenge).decode('utf-8')
        code_challenge = code_challenge.replace('=', '')
        return code_verifier, code_challenge

    def generate_authorize_code_URL(self, challenge, listen_port):
        authorize_code_URL = 'https://accounts.google.com/o/oauth2/auth?response_type=code&client_id='
        authorize_code_URL += self.clientId + '&'
        authorize_code_URL += 'redirect_uri=http://127.0.0.1:' + str(listen_port) + '&'
        authorize_code_URL += 'scope=' + self.scope + '&'
        authorize_code_URL += 'code_challenge=' + challenge + '&code_challenge_method=S256'
        return authorize_code_URL

    def  get_refresh_token_request_payload(self, authorization_code, verifier, listen_port):
        payload = 'grant_type=authorization_code&code=' + authorization_code 
        payload += '&code_verifier=' + verifier
        payload += '&redirect_uri=http://127.0.0.1:' + str(listen_port)
        payload += '&client_id=' + self.clientId
        payload += '&client_secret=' + self.clientSecret
        return payload

    def generate_refresh_token(self):
        #create a Socket
        soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        soc.settimeout(300)
        try:
            soc.bind(('127.0.0.1', 0))

            sock_address = soc.getsockname()
            listen_port = sock_address[1]
            
        except socket.error as message:
            print('Bind failed. Error Code : '
                + str(message[0]) + ' Message '
                + message[1])
            sys.exit()

        #Generate code verifier and challenge
        code_verifier, code_challenge = self.generate_code_verifier_and_challenge()

        # Get the URL to log in for Authorization code.
        authorize_code_URL = self.generate_authorize_code_URL(code_challenge, listen_port)
        print()
        print ('Your local browser has opened the following URL for you to authorize and generate a refresh token:')
        print(authorize_code_URL)

        try:
            # Try to open the browser. 
            if not webbrowser.open(authorize_code_URL):
                print('This script requires a local web browser, signed in with your User Account, to authorize the drivers and generate the refresh token. Please install a web browser or run this script on a machine with a browser and signed in User Account.')
        except webbrowser.Error:
                print('This script requires a local web browser, signed in with your User Account, to authorize the drivers and generate the refresh token. Please install a web browser or run this script on a machine with a browser and signed in User Account.')

        #Listen to the port to get data
        soc.listen()        
        conn, address = soc.accept()

        data = conn.recv(4000)
        if not data:
            print('No data was recieved from login URL.')
            sys.exit()

        successPage = 'HTTP/1.1 303 See Other\nLocation: https://cloud.google.com/bigquery/docs/reference/odbc-jdbc-drivers\n\n'
        conn.send(successPage.encode())
        conn.close()

        #extract authorization code from data that we got on local host.
        authorization_code= self.ExtractCode(data.decode("utf-8"))
        if authorization_code == "" :
            print('Authorization code could not get extracted from recieved data.')
            sys.exit()

        # send authentication request to get refresh token.
        payload = self.get_refresh_token_request_payload(authorization_code, code_verifier, listen_port)
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        httpsConn = http.client.HTTPSConnection("oauth2.googleapis.com")
        httpsConn.request("POST", "/token", payload, headers)
        res = httpsConn.getresponse()
        data = res.read()
        parsedData = json.loads(data.decode('utf-8'))
        if 'refresh_token' not in parsedData:
            print()
            if 'error' in parsedData:
                print('error: ' + parsedData['error'])
            if 'error_description' in parsedData:
                print('description: ' + parsedData['error_description'])
            sys.exit()

        refreshToken = parsedData["refresh_token"]
        if (not refreshToken) or (refreshToken==""):
            print('No refresh token was generated.')
            sys.exit()

        print()
        print('The refresh token was generated. You can now copy it to connection string of driver.')
        print("RefreshToken=" + refreshToken)
        return refreshToken


if __name__ == "__main__": 
    user_auth = UserAccountAuthentication(CLIENT_ID,CLIENT_SECRET,SCOPE)
    refresh_token = user_auth.generate_refresh_token()